# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/core/00_data.ipynb.

# %% auto 0
__all__ = ['BASE_URL', 'fetch_csv', 'Collins', 'Kras']

# %% ../../nbs/core/00_data.ipynb 3
import pandas as pd
import requests
from functools import lru_cache

# %% ../../nbs/core/00_data.ipynb 7
BASE_URL = "https://github.com/sky1ove/kdock/raw/main/"

# %% ../../nbs/core/00_data.ipynb 8
@lru_cache()
def fetch_csv(url):
    return pd.read_csv(url)

# %% ../../nbs/core/00_data.ipynb 9
class Collins:
    "A class of loading compound datasets from Collins lab."

    @staticmethod
    def get_antibiotics_2k():
        """
        Antibiotics dataset of 50 µM 2,560 compounds screening in E. coli K12 BW25113.
        2,335 unique compounds after deduplicated.
        Table S1B from 2020 Cell: A Deep Learning Approach to Antibiotic Discovery.
        """
        return fetch_csv(BASE_URL + "dataset/antibiotics_2k.csv")

    @staticmethod
    def get_antibiotics_39k():
        """
        Antibiotics dataset of 50 µM 39,128 compounds screening in E. coli K12 BW25113.
        Supplementary dataset EV1 from 2022 Molecular Systems Biology: Benchmarking AlphaFold-enabled molecular docking predictions for antibiotic discovery.
        """
        return fetch_csv(BASE_URL + "dataset/antibiotics_39k.csv")

    @staticmethod
    def get_antibiotics_enzyme():
        """
        Antibiotics enzymatic inhibition dataset of 100 µM 218 compounds and 12 essential proteins in E. coli K12 BW25113.
        Flattened benchmark dataset/Supplementary EV4 from 2022 Molecular Systems Biology: Benchmarking AlphaFold-enabled molecular docking predictions for antibiotic discovery.
        """
        return fetch_csv(BASE_URL + "dataset/antibiotics_enzyme.csv")

# %% ../../nbs/core/00_data.ipynb 19
class Kras:
    "A class of fetching various KRAS datasets."
    @staticmethod
    def get_mirati_g12d():
        "Deduplicated G12D dataset from the mirati paper and patents."
        return fetch_csv(BASE_URL + "dataset/KRASi_g12d_dedup.csv")

    @staticmethod
    def get_mirati_g12d_raw():
        "Raw G12D dataset from the paper and patents without deduplication."
        return fetch_csv(BASE_URL + "dataset/KRASi_g12d.csv")

    @staticmethod
    def get_seq():
        "Protein sequence of human KRAS and its mutants G12D and G12C."
        return fetch_csv(BASE_URL + "dataset/kras_seq.csv")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/core/02_protein.ipynb.

# %% auto 0
__all__ = ['get_uniprot_seq', 'get_uniprot_features', 'get_uniprot_kd', 'get_uniprot_type', 'apply_mut_single',
           'apply_mut_complex', 'compare_seq']

# %% ../../nbs/core/02_protein.ipynb 3
import pandas as pd
import requests,re
from functools import lru_cache

# for compare seq
from Bio.Align import PairwiseAligner

# %% ../../nbs/core/02_protein.ipynb 5
@lru_cache()
def get_uniprot_seq(uniprot_id):
    "Queries the UniProt database to retrieve the protein sequence for a given UniProt ID."
    
    url = f"https://www.uniprot.org/uniprot/{uniprot_id}.fasta"
    response = requests.get(url)

    # Check if the request was successful (status code 200)
    if response.status_code == 200:
        data = response.text
        # The sequence starts after the first line, which is a description
        sequence = ''.join(data.split('\n')[1:]).strip()
        return sequence
    else:
        return f"Error: Unable to retrieve sequence for UniProt ID {uniprot_id}. Status code: {response.status_code}"

# %% ../../nbs/core/02_protein.ipynb 7
@lru_cache()
def get_uniprot_features(uniprot_id):
    "Given uniprot_id, get specific region for uniprot features."
    # uniprot REST API
    url = f"https://rest.uniprot.org/uniprotkb/{uniprot_id}.json"
    response = requests.get(url)

    if response.status_code == 200:
        data = response.json()
        # Extract the "features" section which contains information
        features = data.get('features', [])
        
        protein_name = (
            data.get("proteinDescription", {})
                .get("recommendedName", {})
                .get("fullName", {})
                .get("value")
        )

        gene_name = (
            data.get("genes", [{}])[0]
                .get("geneName", {})
                .get("value")
        )
        return {
            "uniprot_id": uniprot_id,
            "protein_name": protein_name,
            "gene_name": gene_name,
            "features": features
        }
    else:
        raise ValueError(f"Failed to retrieve UniProt features for {uniprot_id}")

# %% ../../nbs/core/02_protein.ipynb 9
def get_uniprot_kd(uniprot_id):
    "Query 'Domain: Protein kinase' based on UniProt ID and get its sequence info."
    data = get_uniprot_features(uniprot_id)
    seq = get_uniprot_seq(uniprot_id)
    out = []

    for feature in data['features']:
        if feature.get("type") == "Domain" and "Protein kinase" in feature.get("description", ""):
            start = feature['location']['start']['value']
            end = feature['location']['end']['value']
            out.append({
                "uniprot_id": uniprot_id,
                "protein_name": data["protein_name"],
                "gene_name": data["gene_name"],
                "start": start,
                "end": end,
                "description": feature.get("description", ""),
                "sequence": seq[start-1:end]
            })

    return out

# %% ../../nbs/core/02_protein.ipynb 11
def get_uniprot_type(uniprot_id,type_='Signal'):
    "Get region sequences based on UniProt ID features."
    data = get_uniprot_features(uniprot_id)
    seq = get_uniprot_seq(uniprot_id)

    out = []

    for feature in data['features']:
        if feature.get("type") == type_:
            start = feature['location']['start']['value']
            end = feature['location']['end']['value']
            region = {
                'uniprot_id': uniprot_id,
                'type': feature['type'],
                "protein_name": data["protein_name"],
                "gene_name": data["gene_name"],
                'start': start,
                'end': end,
                'description': feature['description'],
                'sequence': seq[start-1:end]
            }
            out.append(region)

    if not out:
        available = sorted({f.get("type") for f in data['features'] if f.get("type")})
        print(f"No feature of type '{type_}' found for {uniprot_id}.")
        print(f"Available feature types: {', '.join(available)}")
        return available

    return out

# %% ../../nbs/core/02_protein.ipynb 16
def apply_mut_single(seq, # protein sequence
           *mutations, # e.g., E709A
           start_pos=1, # if the protein sequence does not start from index 1, indicate the start index to match the mutations
           ):
    "Apply mutations to a protein sequence."
    seq_list = list(seq)  # convert to list for mutability
    
    for mut in mutations:
        # check mutation format
        if len(mut) < 3: raise ValueError(f"Invalid mutation format: {mut}")
        
        from_aa,pos,to_aa = mut[0],int(mut[1:-1])-start_pos,mut[-1]

        # make sure position is within the sequence length
        if pos < 0 or pos >= len(seq_list): raise IndexError(f"Position {pos + 1} out of range for sequence length {len(seq_list)}")
        # make sure aa from mutations matches the residue on the sequence
        if seq_list[pos] != from_aa: raise ValueError(f"Expected {from_aa} at position {pos + 1}, found {seq_list[pos]}")
        
        seq_list[pos] = to_aa
        print('Converted:', mut)
        
    return ''.join(seq_list)

# %% ../../nbs/core/02_protein.ipynb 18
def apply_mut_complex(seq, # protein sequence
                      mut, # mutation (e.g., G776delinsVC/S783C, G778dupGSP)
                      start_pos=1, # if truncated protein sequence, indicate where it starts to match the position of mutation
                      ):
    """
    Apply a composite mutation like 'G776delinsVC/S783C' to `seq`,
    assuming `seq[0]` corresponds to residue number `start_pos`.

    * At most one delins **or** dup is allowed.
    * Point substitutions are executed first; the indel/dup is done last.
    """
    _sub_pat    = re.compile(r'^([A-Z])(\d+)([A-Z])$')          # e.g. S783C
    _delins_pat = re.compile(r'^([A-Z])(\d+)delins([A-Z]+)$')   # e.g. G776delinsVC
    _dup_pat    = re.compile(r'^([A-Z])(\d+)dup([A-Z]+)$')      # e.g. G778dupGSP

    seq = list(seq)
    tokens = mut.split('/')

    # ---------- 1) substitutions (length-neutral) ----------
    for m in tokens:
        if _sub_pat.match(m):
            orig, pos, new = _sub_pat.match(m).groups()
            idx = int(pos) - start_pos
            if seq[idx] != orig:
                raise ValueError(
                    f"Mismatch at position {pos}: expected {orig}, got {seq[idx]}"
                )
            seq[idx] = new

    # ---------- 2) the single length-changing event ----------
    for m in tokens:
        if _delins_pat.match(m):
            orig, pos, ins = _delins_pat.match(m).groups()
            idx = int(pos) - start_pos
            if seq[idx] != orig:
                raise ValueError(
                    f"Mismatch at position {pos}: expected {orig}, got {seq[idx]}"
                )
            seq[idx : idx + 1] = list(ins)         # replace 1 residue with many
            break

        if _dup_pat.match(m):
            orig, pos, dup = _dup_pat.match(m).groups()
            idx = int(pos) - start_pos
            if seq[idx] != orig:
                raise ValueError(
                    f"Mismatch at position {pos}: expected {orig}, got {seq[idx]}"
                )
            seq[idx + 1 : idx + 1] = list(dup)     # insert right after the residue
            break

    return ''.join(seq)

# %% ../../nbs/core/02_protein.ipynb 22
def compare_seq(
    seq1: str,  # original
    seq2: str,  # mutant
    *,
    start_pos: int = 1,
    label1: str = "Original",
    label2: str = "Mutant",
    visualize: bool = True
):
    """
    Align two protein sequences and summarise differences.
    """

    # ----- global alignment using PairwiseAligner -----
    aligner = PairwiseAligner()
    aligner.mode = "global"
    aligner.match_score = 2
    aligner.mismatch_score = -1
    aligner.open_gap_score = -5
    aligner.extend_gap_score = -0.5

    alignment = aligner.align(seq1, seq2)[0]
    aln1 = alignment.aligned[0]
    aln2 = alignment.aligned[1]

    # Reconstruct aligned strings from aligned segments
    aligned_seq1 = []
    aligned_seq2 = []
    i1, i2 = 0, 0

    for (start1, end1), (start2, end2) in zip(aln1, aln2):
        # Add gaps to make sequences align
        while i1 < start1:
            aligned_seq1.append(seq1[i1])
            aligned_seq2.append('-')
            i1 += 1
        while i2 < start2:
            aligned_seq1.append('-')
            aligned_seq2.append(seq2[i2])
            i2 += 1

        # Add aligned part
        for j in range(end1 - start1):
            aligned_seq1.append(seq1[i1])
            aligned_seq2.append(seq2[i2])
            i1 += 1
            i2 += 1

    # Remaining tails
    while i1 < len(seq1):
        aligned_seq1.append(seq1[i1])
        aligned_seq2.append('-')
        i1 += 1
    while i2 < len(seq2):
        aligned_seq1.append('-')
        aligned_seq2.append(seq2[i2])
        i2 += 1

    aln1_str = ''.join(aligned_seq1)
    aln2_str = ''.join(aligned_seq2)

    # ----- find differences -----
    diffs, raw_i1, raw_i2 = [], 0, 0
    for a1, a2 in zip(aln1_str, aln2_str):
        if a1 != '-' and a2 != '-':
            if a1 != a2:
                diffs.append((start_pos + raw_i1, a1, a2, 'substitution'))
            raw_i1 += 1
            raw_i2 += 1
        elif a1 == '-' and a2 != '-':
            diffs.append((start_pos + raw_i1, '-', a2, 'insertion'))
            raw_i2 += 1
        elif a1 != '-' and a2 == '-':
            diffs.append((start_pos + raw_i1, a1, '-', 'deletion'))
            raw_i1 += 1

    # ----- visualization -----
    if visualize:
        for block in range(0, len(aln1_str), 80):
            s1_block = aln1_str[block:block + 80]
            s2_block = aln2_str[block:block + 80]
            marker = ''.join(' ' if x == y else '^' for x, y in zip(s1_block, s2_block))
            left_idx = start_pos + aln1_str[:block].replace('-', '').__len__()
            right_idx = left_idx + s1_block.replace('-', '').__len__() - 1
            print(f"{label1:<10} {left_idx:>5}-{right_idx:<5}: {s1_block}")
            print(f"{label2:<10} {'':>11}: {s2_block}")
            print(f"{'':>22}  {marker}\n")

    # ----- summary list -----
    print("Differences:")
    for pos, ref, new, kind in diffs:
        print(f"  {kind:<12} at {pos:>4}: {ref} → {new}")

    # return diffs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/core/03_ligand.ipynb.

# %% auto 0
__all__ = ['plot_drug', 'rdkit_conformer', 'tanimoto', 'get_rdkit', 'get_rdkit_3d', 'get_rdkit_all', 'remove_hi_corr',
           'preprocess', 'get_rdkit_df', 'get_fp', 'compress_fp', 'tanimoto_numba', 'hash_fp', 'get_same_mol_group']

# %% ../../nbs/core/03_ligand.ipynb 3
import pandas as pd, numpy as np
from rdkit import Chem,RDLogger,DataStructs
from rdkit.Chem import Draw, rdDepictor, Descriptors,Descriptors3D, AllChem
from skfp.fingerprints import AtomPairFingerprint, ECFPFingerprint, MACCSFingerprint
import hashlib,numba
from .data import *

from tqdm.contrib.concurrent import process_map
from tqdm import tqdm
RDLogger.DisableLog('rdApp.warning')

# %% ../../nbs/core/03_ligand.ipynb 5
def plot_drug(drug_dict, flip_list=None, save_path=None):
    flip_list = flip_list or []
    mols = []

    for name, smi in drug_dict.items():
        mol = Chem.MolFromSmiles(smi)
        rdDepictor.Compute2DCoords(mol)

        # Flip horizontally if requested
        if name in flip_list:
            conf = mol.GetConformer()
            for i in range(conf.GetNumAtoms()):
                x, y, z = conf.GetAtomPosition(i)
                conf.SetAtomPosition(i, (-x, y, z))

        mol.SetProp("_Name", name)
        mols.append(mol)

    # Draw to SVG
    svg_obj = Draw.MolsToGridImage(
        mols,
        molsPerRow=3,
        subImgSize=(300, 250),
        legends=[m.GetProp("_Name") for m in mols],
        useSVG=True
    )

    svg_str = getattr(svg_obj, "data", getattr(svg_obj, "GetDrawingText", lambda: svg_obj)())

    if save_path:
        with open(save_path, "w", encoding="utf-8") as f:
            f.write(svg_str)
        print(f"✅ Saved structure grid to: {save_path}")
    else:
        # from IPython.display import SVG, display
        # display(SVG(svg_str))
        return svg_obj

# %% ../../nbs/core/03_ligand.ipynb 8
def rdkit_conformer(SMILES, # SMILES string
                    output=None, # file ".sdf" to be saved
                    method='ETKDG', # Optimization method, can be 'UFF', 'MMFF' or 'ETKDGv3'
                    visualize=True, #whether or not to visualize the compound
                    seed = 3,# randomness of the 3D conformation
                    ):

    "Gemerate 3D conformers from SMILES"
    
    np.random.seed(seed) 
    mol = Chem.MolFromSmiles(SMILES)
    
    # Generate a 3D conformation of the molecule
    AllChem.EmbedMolecule(mol)
    

    # Optimize the 3D conformation using the specified force field method
    if method == 'UFF':
        AllChem.UFFOptimizeMolecule(mol)
    elif method == 'MMFF':
        AllChem.MMFFOptimizeMolecule(mol)
    elif method == 'ETKDG':
        AllChem.EmbedMultipleConfs(mol, numConfs=1, useExpTorsionAnglePrefs=True, 
                                   useBasicKnowledge=True, enforceChirality=True,randomSeed=seed)
        AllChem.ETKDGv3()
        AllChem.UFFOptimizeMolecule(mol)

    else:
        raise ValueError('Invalid method specified')
        

    # Remove hydrogens from the molecule
    # mol = Chem.RemoveHs(mol)

    if output is not None:
        Path(output).parent.mkdir(parents=True,exist_ok=True)
    
        w = Chem.SDWriter(output)
        w.write(mol)
        w.close()
    return mol

# %% ../../nbs/core/03_ligand.ipynb 10
def tanimoto(df, # df with SMILES and ID columns
             smiles_col='SMILES', # colname of SMILES
             id_col='ID', # colname of compound ID
             target_col=None, # colname of compound values (e.g., IC50)
             radius=2, # radius of the Morgan fingerprint.
             ):
    "Calculates the Tanimoto similarity scores between all pairs of molecules in a pandas DataFrame."
    
    df = df.copy()
    # Convert SMILES to molecule objects
    df['Molecule'] = df[smiles_col].apply(lambda x: Chem.MolFromSmiles(x))

    # Calculate fingerprints
    df['Fingerprint'] = df['Molecule'].apply(lambda x: AllChem.GetMorganFingerprintAsBitVect(x, radius))

    # Calculate similarity scores
    similarity_scores = []
    for i in range(len(df)):
        for j in range(i+1, len(df)):
            sim_score = DataStructs.TanimotoSimilarity(df['Fingerprint'][i], df['Fingerprint'][j])
            if target_col is not None:
                similarity_scores.append((df[id_col][i], df[id_col][j], df[smiles_col][i], df[smiles_col][j], sim_score, df[target_col][i], df[target_col][j]))
            else:
                similarity_scores.append((df[id_col][i], df[id_col][j], df[smiles_col][i], df[smiles_col][j], sim_score))

    # Create a new DataFrame with the similarity scores
    if target_col is not None:
        result_df = pd.DataFrame(similarity_scores, columns=['ID1', 'ID2', 'SMILES1', 'SMILES2', 'SimilarityScore', 'Target1', 'Target2'])
    else:
        result_df = pd.DataFrame(similarity_scores, columns=['ID1', 'ID2', 'SMILES1', 'SMILES2', 'SimilarityScore'])

    # Sort by similarity score in descending order
    result_df.sort_values('SimilarityScore', ascending=False, inplace=True)
    result_df = result_df.reset_index(drop=True)

    return result_df

# %% ../../nbs/core/03_ligand.ipynb 14
def get_rdkit(SMILES:str):
    """
    Extract chemical features from SMILES
    Reference: https://greglandrum.github.io/rdkit-blog/posts/2022-12-23-descriptor-tutorial.html
    """
    mol = Chem.MolFromSmiles(SMILES)
    return Descriptors.CalcMolDescriptors(mol)

# %% ../../nbs/core/03_ligand.ipynb 15
def get_rdkit_3d(SMILES:str):
    "Extract 3d features from SMILES"
    mol = Chem.MolFromSmiles(SMILES)
    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol, AllChem.ETKDG())
    AllChem.UFFOptimizeMolecule(mol)
    return Descriptors3D.CalcMolDescriptors3D(mol)

# %% ../../nbs/core/03_ligand.ipynb 16
def get_rdkit_all(SMILES:str):
    "Extract chemical features and 3d features from SMILES"
    feat = get_rdkit(SMILES)
    feat_3d = get_rdkit_3d(SMILES)
    return feat|feat_3d

# %% ../../nbs/core/03_ligand.ipynb 17
def remove_hi_corr(df: pd.DataFrame, 
                   thr=0.99 # threshold
                   ):
    "Remove highly correlated features in a dataframe given a pearson threshold"
    corr_matrix = df.corr().abs()
    upper = corr_matrix.where(np.triu(np.ones(corr_matrix.shape), k=1).astype(bool))
    to_drop = [column for column in upper.columns if any(upper[column] > thr)]
    return df.drop(to_drop, axis=1), to_drop

# %% ../../nbs/core/03_ligand.ipynb 18
def preprocess(df: pd.DataFrame, thr=0.99):
    "Remove features with no variance, and highly correlated features based on threshold."
    col_ori = df.columns

    # Remove columns with std == 0
    std_zero_cols = df.columns[df.std() == 0].tolist()
    
    if std_zero_cols:
        n=len(std_zero_cols)
        print(f"\n {n} Columns with zero std: {std_zero_cols}")
    df = df.loc[:, df.std() != 0].copy()

    # Remove highly correlated columns
    df, high_corr_cols = remove_hi_corr(df, thr)
    if high_corr_cols:
        n=len(high_corr_cols)
        print(f"\n {n} Columns removed due to high similarity (pearson>{thr}): {high_corr_cols}")

    dropping_col = set(col_ori) - set(df.columns)
    n = len(dropping_col)
    print(f"\n Total removed columns: {n}")
    
    return df

# %% ../../nbs/core/03_ligand.ipynb 19
def get_rdkit_df(df: pd.DataFrame,
                 include_3d=False,
                 col='SMILES',
                 postprocess=False,
                 chunksize=128, # for parallel process_map 
                 ):
    "Extract rdkit features (optionally in parallel with 3D) from SMILES in a df"
    func = get_rdkit_all if include_3d else get_rdkit
    smiles_list = df[col].tolist()

    features = process_map(func, smiles_list,chunksize=chunksize)

    out = pd.DataFrame(features)

    if postprocess:
        out = StandardScaler().fit_transform(out)
        out = preprocess(out) # remove redundant

    return out

# %% ../../nbs/core/03_ligand.ipynb 25
def get_fp(SMILES, 
           name='ecfp',
           ELEMENTS_PER_WORKER = 1_000_000):
    "Super fast method to get molecule fingerprints using scikit-fingerprints"
    if name == 'ecfp':
        fp_transformer = ECFPFingerprint(fp_size=2048, radius=3, n_jobs=-1)
    elif name == 'ap':
        fp_transformer = AtomPairFingerprint(fp_size=2048, n_jobs=-1)
    elif name == 'maccs':
        fp_transformer = MACCSFingerprint(n_jobs=-1)
    else:
        raise Exception('Wrong fingerprint name!')
    
    middle_parts = []
    k_splits = len(SMILES) // ELEMENTS_PER_WORKER
        
    for i in tqdm(range(k_splits)):
        middle_parts.append(fp_transformer.transform(SMILES[i * ELEMENTS_PER_WORKER: (i + 1) * ELEMENTS_PER_WORKER]))
        
    if len(SMILES) % ELEMENTS_PER_WORKER > 0:   
        middle_parts.append(fp_transformer.transform(SMILES[k_splits * ELEMENTS_PER_WORKER:]))
    
    return np.concatenate(middle_parts)

# %% ../../nbs/core/03_ligand.ipynb 29
def compress_fp(array):
    "Compress binary finterprints using np.packbits"
    return np.packbits(array,axis=1)

# %% ../../nbs/core/03_ligand.ipynb 47
@numba.njit(parallel=True)
def tanimoto_numba(fps):
    "Get a NxN matrix of tanimoto similarity among N compounds."
    n = fps.shape[0]
    result = np.zeros((n, n), dtype=np.float32)
    
    for i in numba.prange(n):
        for j in range(i, n):  # only upper triangle
            inter = np.bitwise_and(fps[i], fps[j]).sum()
            union = np.bitwise_or(fps[i], fps[j]).sum()
            sim = inter / union if union > 0 else 0.0
            result[i, j] = sim
            result[j, i] = sim  # fill symmetric position
    return result

# %% ../../nbs/core/03_ligand.ipynb 50
def hash_fp(fp_row):
    "Hash a binary fingerprint row using SHA256"
    return hashlib.sha256(fp_row.tobytes()).hexdigest()

# %% ../../nbs/core/03_ligand.ipynb 52
def get_same_mol_group(df, smi_col='SMILES'):
    "Assign a group number to the same compounds by utilizing hash sha256 to encode morgan fp and find same molecule."
    df = df.copy()
    smiles = df[smi_col].tolist()
    
    fps = get_fp(smiles)
    
    # Hash each fingerprint
    # hashes = [hash_fp(fp) for fp in fps] # non-parallel
    hashes = process_map(hash_fp, fps, chunksize=256) # if parallel
    
    # Assign a group number based on hash buckets
    hash_to_group = {}
    group_ids = []
    group_counter = 0
    for h in hashes:
        if h not in hash_to_group:
            hash_to_group[h] = group_counter
            group_counter += 1
        group_ids.append(hash_to_group[h])
    
    df['group'] = group_ids
    return df



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/core/01_utils.ipynb.

# %% auto 0
__all__ = ['rglob', 'copy_files', 'get_rec_lig', 'get_box', 'view_mol', 'view_complex']

# %% ../../nbs/core/01_utils.ipynb 3
from pathlib import Path
import subprocess,shutil,zipfile
import numpy as np

import py3Dmol
from rdkit import Chem

# %% ../../nbs/core/01_utils.ipynb 6
def rglob(path, pattern, max_depth):
    "Get a file list given folder depths"
    base_path = Path(path).resolve()
    for path in base_path.rglob(pattern):
        if len(path.relative_to(base_path).parts) <= max_depth:
            yield path

# %% ../../nbs/core/01_utils.ipynb 8
def copy_files(file_list, dest_dir):
    "Copy a list of files to the destination directory, or zip them if dest_dir ends with .zip."
    dest_path = Path(dest_dir)

    if dest_path.suffix == ".zip":
        with zipfile.ZipFile(dest_path, 'w') as zipf:
            for file_path in file_list:
                file_path = Path(file_path)
                zipf.write(file_path, arcname=file_path.name)
        print(f'Zipped {len(file_list)} files to {dest_path}')
    else:
        dest_path.mkdir(parents=True, exist_ok=True)
        for file_path in file_list:
            file_path = Path(file_path)
            shutil.copy2(file_path, dest_path / file_path.name)
        print(f'Copied {len(file_list)} files to {dest_path}')

# %% ../../nbs/core/01_utils.ipynb 11
def get_rec_lig(pdb_id: str, # pdb id for download
                            lig_id: str, # ligand id shown on the protein page
                            out_dir = '.', # directory path to save pdb files
                            ):
    "Download pdb and extract receptor and ligand from a PDB ID."
    out_dir = Path(out_dir).expanduser().resolve()
    out_dir.mkdir(parents=True, exist_ok=True)

    pdb_file = out_dir / f"{pdb_id}.pdb"
    rec_file = out_dir / f"{pdb_id}_receptor.pdb"
    lig_pdb_file = out_dir / f"{pdb_id}_lig.pdb"
    lig_sdf_file = out_dir / f"{pdb_id}_lig.sdf"

    # Download if not exists
    if not pdb_file.exists():
        url = f"http://files.rcsb.org/download/{pdb_id}.pdb"
        print(f'Downloading pdb: {pdb_id}')
        subprocess.run(["wget", url, "-O", str(pdb_file)], check=True,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print(f'{pdb_id}.pdb is detected!')

    # Extract protein (all ATOM lines excluding ligand ID)
    with open(pdb_file) as infile, open(rec_file, 'w') as out_rec:
        for line in infile:
            if line.startswith("ATOM") and lig_id not in line:
                out_rec.write(line)

    # Extract ligand
    with open(pdb_file) as infile, open(lig_pdb_file, 'w') as out_lig:
        for line in infile:
            if lig_id in line and line.startswith(("HETATM", "ATOM")):
                out_lig.write(line)

    # Convert ligand PDB to SDF using RDKit
    mol = Chem.MolFromPDBFile(str(lig_pdb_file), removeHs=False)
    if mol is None:
        raise ValueError("Failed to parse ligand from PDB.")
    
    writer = Chem.SDWriter(str(lig_sdf_file))
    writer.write(mol)
    writer.close()

    return str(rec_file), str(lig_sdf_file)

# %% ../../nbs/core/01_utils.ipynb 14
def get_box(sdf_file, autobox_add=4.0,tolist=False):
    "Get the box coordinates of ligand.sdf; mimic GNINA's --autobox_ligand behavior."
    mol = Chem.SDMolSupplier(str(sdf_file), removeHs=False)[0]
    if mol is None:
        raise ValueError(f"Failed to read molecule from {sdf_file}")
    
    conf = mol.GetConformer()
    coords = np.array([list(conf.GetAtomPosition(i)) for i in range(mol.GetNumAtoms())])
    
    min_coords = coords.min(axis=0)
    max_coords = coords.max(axis=0)
    
    center = (min_coords + max_coords) / 2
    size = (max_coords - min_coords) + autobox_add

    box_dict = {
        "center_x": round(float(center[0]), 3),
        "center_y": round(float(center[1]), 3),
        "center_z": round(float(center[2]), 3),
        "size_x": round(float(size[0]), 3),
        "size_y": round(float(size[1]), 3),
        "size_z": round(float(size[2]), 3)
    }
    return list(box_dict.values()) if tolist else box_dict

# %% ../../nbs/core/01_utils.ipynb 18
def view_mol(file, #sdf or pdb file
            ):
    "Visualize pdb or sdf file"
    
    v = py3Dmol.view()
    v.addModel(open(file).read())
    v.setStyle({'stick':{}})
    v.zoomTo()
    v.show()

# %% ../../nbs/core/01_utils.ipynb 20
def view_complex(receptor,           # protein file
                 ligand,             # ligand (green), or docked ligand
                 ori_ligand=None,    # original ligand (yellow)
                 box=None            # optional box: [x, y, z, sizeX, sizeY, sizeZ]
                ):

    "Visualize the receptor, ligand, optional original ligand, and optional box via py3Dmol."
    v = py3Dmol.view()
    
    # Load receptor
    v.addModel(open(receptor).read())
    v.setStyle({'cartoon': {}, 'stick': {'radius': 0.15}})
    
    # Load docked ligand
    v.addModel(open(ligand).read())
    v.setStyle({'model': 1}, {'stick': {'colorscheme': 'greenCarbon'}})

    # Load original ligand if provided
    if ori_ligand is not None:
        v.addModel(open(ori_ligand).read())
        v.setStyle({'model': 2}, {'stick': {'colorscheme': 'yellowCarbon'}})

    # Add bounding box if specified
    if box is not None and len(box) == 6:
        x, y, z, sizeX, sizeY, sizeZ = box
        v.addBox({
            'center': {'x': x, 'y': y, 'z': z},
            'dimensions': {'w': sizeX, 'h': sizeY, 'd': sizeZ},
            'color': 'red',
            'opacity': 1,
            'wireframe': True
        })

    v.zoomTo({'model': 1})
    v.show()



"""Functions to plot on 2D"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/core/04_plot.ipynb.

# %% auto 0
__all__ = ['reduce_feature', 'set_sns', 'plot_2d', 'plot_corr']

# %% ../../nbs/core/04_plot.ipynb 3
import pandas as pd, seaborn as sns
from fastcore.meta import delegates
from matplotlib import pyplot as plt

# # kdock
# from kdock.core.data import *

# dimentional reduction
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
from umap import UMAP

# %% ../../nbs/core/04_plot.ipynb 6
def reduce_feature(data, # df or numpy array
                   method='pca', # dimensionality reduction method, accept both capital and lower case
                   complexity=20, # None for PCA; perfplexity for TSNE, recommend: 30; n_neigbors for UMAP, recommend: 15
                   n=2, # n_components
                   seed: int=123, # seed for random_state
                   **kwargs, # arguments from PCA, TSNE, or UMAP depends on which method to use
                  ):
    
    "Reduce the dimensionality given a dataframe of values"
    
    method = method.lower()
    assert method in ['pca','tsne','umap'], "Please choose a method among PCA, TSNE, and UMAP"

    if method == 'pca':
        reducer = PCA(n_components=n, random_state=seed,**kwargs)
    elif method == 'tsne':
        reducer = TSNE(n_components=n,
                       random_state=seed, 
                       perplexity = complexity, # default from official is 30 
                      **kwargs)
    elif method == 'umap':
        reducer = UMAP(n_components=n, 
                       random_state=seed, 
                       n_neighbors=complexity, # default from official is 15, try 15-200
                      **kwargs)
    else:
        raise ValueError('Invalid method specified')

    proj = reducer.fit_transform(data)
    embedding_df = pd.DataFrame(proj).set_index(data.index) if isinstance(data,pd.DataFrame) else pd.DataFrame(proj)
    embedding_df.columns = [f"{method.upper()}{i}" for i in range(1, n + 1)]

    return embedding_df

# %% ../../nbs/core/04_plot.ipynb 10
def set_sns():
    sns.set(rc={"figure.dpi":300, 'savefig.dpi':300})
    sns.set_context('notebook')
    sns.set_style("ticks")

# %% ../../nbs/core/04_plot.ipynb 12
@delegates(sns.scatterplot)
def plot_2d(X: pd.DataFrame, # a dataframe that has first column to be x, and second column to be y
            **kwargs, # arguments for sns.scatterplot
            ):
    "Make 2D plot from a dataframe that has first column to be x, and second column to be y"
    plt.figure(figsize=(7,7))
    sns.scatterplot(data = X,x=X.columns[0],y=X.columns[1],alpha=0.7,**kwargs)

# %% ../../nbs/core/04_plot.ipynb 14
def plot_corr(x,#a column of df
              y,#a column of df
              xlabel=None,# x axis label
              ylabel=None,# y axis label
              order=3, # polynomial level, if straight, order=1 
             ):
    sns.regplot(x=x,
            y=y,
            order=order,
            line_kws={'color': 'gray'}
           )
    
    if xlabel is not None:
        plt.xlabel(xlabel)
        
    if ylabel is not None:
        plt.ylabel(ylabel)
    # plt.text(x=0.8, y=0.1, s=f'Spearman: {correlation:.2f}', transform=plt.gca().transAxes, ha='center', va='center');



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/af3/01_docker.ipynb.

# %% auto 0
__all__ = ['get_docker_command', 'docker_single_full', 'docker_multi_full', 'docker_multi_msa', 'docker_multi_infer']

# %% ../../nbs/af3/01_docker.ipynb 3
from fastcore.meta import delegates

# %% ../../nbs/af3/01_docker.ipynb 6
def get_docker_command(
    input_dir="af_input", # Directory for input data
    output_dir="af_output", # Directory for output results
    model_dir="af_model", # Directory containing models
    db_dir="af_database", # Directory for databases. If None, this option is ommitted
    cache_dir="af_cache", # Directory for JAX compilation cache. If None, this option is omitted
    gpus=0, # GPU devices to allocate (e.g., 0,1), if None, ommitted
    docker_name="sky1ove/alphafold3", # Docker image name
    embedding=False, # If True, includes the --save_embeddings=true flag
    skip_search=False, # if MSA is precalculated and present in json; If True, includes the --norun_data_pipeline flag
    search_only=False, # search MSA only; If True, sets skip_search to False and includes the --norun_inference flag
    json_path=None  # Path to JSON file. If not None, uses json_file instead of input_dir
):

    "Generate a Docker run command for Alphafold with customizable parameters."
    
    # Start building the Docker command
    cmd_parts = ["docker run --rm \\"]

    if json_path:
        parent_dir=json_path.split('/')[0]
        cmd_parts.append(f'    --volume "$HOME/{parent_dir}:/root/af_input" \\')
    else:
        parent_dir = input_dir.split('/')[0]
        cmd_parts.append(f'    --volume "$HOME/{parent_dir}:/root/af_input" \\')

    cmd_parts.extend([f'    --volume "$HOME/{output_dir}:/root/af_output" \\',
                      f'    --volume "$HOME/{model_dir}:/root/models" \\',
                     ])
    if db_dir:
        cmd_parts.append(f'    --volume "$HOME/{db_dir}:/root/public_databases" \\')

    if cache_dir:
        cmd_parts.append(f'    --volume "$HOME/{cache_dir}:/root/cache" \\')

    if gpus is not None:
        cmd_parts.append(f'    --gpus "device={gpus}" \\')

    cmd_parts.extend([f'    {docker_name} \\',
                      '    python run_alphafold.py \\'])
    
    if json_path:
        mount_path= '/'.join(json_path.split('/')[1:])
        cmd_parts.append(f'    --json_path=/root/af_input/{mount_path} \\')
    else:
        len_dir = len(input_dir.split('/')[1:])
        if len_dir:
            mount_dir= '/'.join(input_dir.split('/')[1:])
            cmd_parts.append(f'    --input_dir=/root/af_input/{mount_dir} \\')
        else:
            cmd_parts.append('    --input_dir=/root/af_input \\')

    cmd_parts.extend(['    --output_dir=/root/af_output \\',
                      '    --model_dir=/root/models \\'])
    
    if cache_dir:
        cmd_parts.append('    --jax_compilation_cache_dir=/root/cache \\')

    if embedding:
        cmd_parts.append('    --save_embeddings=true \\')
    
    if search_only:
        skip_search = False
        cmd_parts.append('    --norun_inference \\')
    elif skip_search:
        cmd_parts.append('    --norun_data_pipeline \\')

    if cmd_parts[-1].endswith('\\'):
        cmd_parts[-1] = cmd_parts[-1].rstrip(' \\')

    docker_command = "\n".join(cmd_parts)
    print(docker_command)

# %% ../../nbs/af3/01_docker.ipynb 12
@delegates(get_docker_command)
def docker_single_full(json_path,output_dir,cache_dir=False,**kwargs):
    "Single json task with full pipeline."
    return get_docker_command(
        json_path=json_path,
        output_dir=output_dir,
        cache_dir=cache_dir,
        **kwargs)

# %% ../../nbs/af3/01_docker.ipynb 16
@delegates(get_docker_command)
def docker_multi_full(input_dir,output_dir,**kwargs):
    "Folder of json as input with full pipeline."
    return get_docker_command(
        input_dir=input_dir,
        output_dir=output_dir,
        **kwargs)

# %% ../../nbs/af3/01_docker.ipynb 19
@delegates(get_docker_command)
def docker_multi_msa(input_dir,output_dir,search_only=True,**kwargs):
    "MSA search only, without structure inference; CPU only."
    return get_docker_command(
        input_dir=input_dir,
        output_dir=output_dir,
        search_only=search_only,
        **kwargs)

# %% ../../nbs/af3/01_docker.ipynb 22
@delegates(get_docker_command)
def docker_multi_infer(input_dir,output_dir,skip_search=True,**kwargs):
    "Infer only with pre-calculated MSA; GPU is needed."
    return get_docker_command(
        input_dir=input_dir,
        output_dir=output_dir,
        skip_search=skip_search,
        **kwargs)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/protenix/06_proteinix_dock.ipynb.

# %% auto 0
__all__ = ['capture_output', 'get_protenix_dock', 'get_protenix_vina_dock', 'json2sdf']

# %% ../../nbs/protenix/06_proteinix_dock.ipynb 11
# kdock
from ..core.utils import *

# basics
import json,shutil
import os,sys,logging,contextlib
from datetime import datetime
from pathlib import Path

# import protenix-dock
try:
    from pxdock import ProtenixDock
    from pxdock.common.rmsd_calculator import write_ligand_to_sdf

    # Prevent multiprocessing-related errors in pxdock
    import multiprocessing as mp
    mp.set_start_method("spawn", force=True)

except ImportError as e:
    print(f"pxdock import failed: {e}")
    ProtenixDock = None
    write_ligand_to_sdf = None


# %% ../../nbs/protenix/06_proteinix_dock.ipynb 15
@contextlib.contextmanager
def capture_output(log_path):
    with open(log_path, 'a') as log_file:  # Append mode prevents overwriting
        # Redirect stdout/stderr
        old_stdout_fd = os.dup(1)
        old_stderr_fd = os.dup(2)
        os.dup2(log_file.fileno(), 1)
        os.dup2(log_file.fileno(), 2)

        # Setup Python logger
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        logger.handlers = [h for h in logger.handlers if not isinstance(h, logging.StreamHandler)]

        file_handler = logging.FileHandler(log_path)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        try:
            yield
        finally:
            os.dup2(old_stdout_fd, 1)
            os.dup2(old_stderr_fd, 2)
            os.close(old_stdout_fd)
            os.close(old_stderr_fd)

# %% ../../nbs/protenix/06_proteinix_dock.ipynb 16
def get_protenix_dock(receptor_pdb, # pdb path
                      ligand_sdf, # sdf path
                      box:list, # center xyz + size xyz
                      out_dir='.', # output directory
                      num_walker=20, # number of Monte Carlo threads, the more walkers the more space to explore
                      mc_prune_energy_threshold=500, # maximum allowed energy for a pose; if exceeds, skipped
                      include_affinity=True, # include bscore
                      **kwargs):
    "Use protenix-dock to dock ligand sdf to receptor pdb using protenix-dock"
    Path(out_dir).mkdir(parents=True,exist_ok=True)
    
    dock_instance = ProtenixDock(receptor_pdb)
    box_center,box_size=box[:3],box[3:]
    dock_instance.set_box(box_center, box_size)
    
    log_path = Path(out_dir)/f'{datetime.now():%Y%m%d_%H%M%S}.log'
    with capture_output(log_path):
        # cache_dir = dock_instance.generate_cache_maps(spacing=0.175)
        result_dir = dock_instance.run_docking(ligand_sdf,out_dir=out_dir,num_walker=num_walker,mc_prune_energy_threshold=mc_prune_energy_threshold,include_affinity=include_affinity,**kwargs)
        # dock_instance.drop_cache_maps()
    
    print(f'Docking complete. Results in: {result_dir}')
    print('Cache dir:', cache_dir)

# %% ../../nbs/protenix/06_proteinix_dock.ipynb 20
def get_protenix_vina_dock(receptor_pdb, # pdb path
                      ligand_sdf, # sdf path
                      box:list, # center xyz + size xyz
                      out_dir='.', # output directory
                      **kwargs):
    "Dock ligand sdf to receptor pdb using protenix-dock"
    Path(out_dir).mkdir(parents=True,exist_ok=True)
    
    dock_instance = ProtenixDock(receptor_pdb)
    box_center,box_size=box[:3],box[3:]
    dock_instance.set_box(box_center, box_size)
    
    log_path = Path(out_dir)/f'{datetime.now():%Y%m%d_%H%M%S}.log'
    with capture_output(log_path):
        result_dir = dock_instance.run_pose_opt(ligand_sdf,out_dir=out_dir,**kwargs)

    print(f'Docking complete. Results in: {result_dir}')

# %% ../../nbs/protenix/06_proteinix_dock.ipynb 23
def json2sdf(json_path,
             sdf_path=None, # .sdf to be saved
             ):
    with open(json_path, "r") as f:
        json_out_data = json.load(f)

    best_pose = json_out_data["best_pose"]
    best_id = best_pose["index"]
    bscore = best_pose["bscore"]
    best_lig = json_out_data["poses"][best_id]
    xyz=best_lig["ligand"]["xyz"]
    if sdf_path is not None:
        write_ligand_to_sdf(json_out_data["mapped_smiles"], [xyz], sdf_path)
    return {'energy':best_lig["energy"],'pscore':best_lig["pscore"],'bscore':bscore}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/protenix/07_protenix.ipynb.

# %% auto 0
__all__ = ['get_single_job', 'get_single_protein_ligand_json', 'get_protein_ligand_df_json', 'get_virtual_screening_json']

# %% ../../nbs/protenix/07_protenix.ipynb 6
import json
from pathlib import Path

# %% ../../nbs/protenix/07_protenix.ipynb 9
def get_single_job(job_name, protein_seq, msa_dir, SMILES=None,CCD=None):
    "Get protenix json format of protein and ligand."
    
    if SMILES and CCD:
        raise ValueError("Please provide only one of SMILES or CCD, not both.")
    if not SMILES and not CCD:
        raise ValueError("You must provide either SMILES or CCD.")

    ligand_value = SMILES if SMILES else f"CCD_{CCD}"
    
    return {
        "name": job_name,
        "sequences": [
            {
                "proteinChain": {
                    "count": 1,
                    "sequence": protein_seq,
                    "msa": {
                        "precomputed_msa_dir": msa_dir,
                        "pairing_db": "uniref100"
                    }
                }
            },
            {
                "ligand": {
                    "count": 1,
                    "ligand": ligand_value
                }
            }
        ]
    }

# %% ../../nbs/protenix/07_protenix.ipynb 11
def get_single_protein_ligand_json(job_name, 
                                   protein_seq, 
                                   msa_dir, 
                                   SMILES=None, 
                                   CCD=None, 
                                   json_path=None):
    "Generate json input for one protein-ligand job."
    data = [get_single_job(job_name, protein_seq, msa_dir, SMILES=SMILES, CCD=CCD)]

    if json_path:
        save_path = Path(json_path)
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with save_path.open("w") as f:
            json.dump(data, f, indent=4)
        print(f"JSON saved to {save_path}")

    return data

# %% ../../nbs/protenix/07_protenix.ipynb 16
def get_protein_ligand_df_json(df,
                               id_col,
                               seq_col, 
                               msa_col, 
                               smi_col=None, 
                               ccd_col=None, 
                               save_json=None):
    "Get json file of protein and ligand in a dataframe."
    
    if smi_col and ccd_col:
        raise ValueError("Provide only one of smi_col or ccd_col, not both.")
    if not smi_col and not ccd_col:
        raise ValueError("You must provide either smi_col or ccd_col.")

    use_smiles = smi_col is not None

    def build_job(row):
        job_name = row[id_col]
        protein_seq = row[seq_col]
        msa_dir = row[msa_col]
        SMILES = row[smi_col] if use_smiles else None
        CCD = None if use_smiles else row[ccd_col]
        return get_single_job(job_name, protein_seq, msa_dir, SMILES=SMILES, CCD=CCD)

    all_jobs = df.apply(build_job, axis=1).tolist()

    if save_json:
        save_path = Path(save_json)
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with save_path.open("w") as f:
            json.dump(all_jobs, f, indent=4)
        print(f"JSON saved to {save_path}")

    return all_jobs


# %% ../../nbs/protenix/07_protenix.ipynb 19
def get_virtual_screening_json(df, 
                               protein_seq, 
                               msa_dir, 
                               id_col,
                               smi_col=None, 
                               ccd_col=None, 
                               save_json=None):
    "Get json file of single protein against multiple SMILES in a dataframe."
    if smi_col and ccd_col:
        raise ValueError("Provide only one of smi_col or ccd_col, not both.")
    if not smi_col and not ccd_col:
        raise ValueError("You must provide either smi_col or ccd_col.")

    use_smiles = smi_col is not None

    def build_job(row):
        job_name = row[id_col]
        SMILES = row[smi_col] if use_smiles else None
        CCD = None if use_smiles else row[ccd_col]
        return get_single_job(job_name, protein_seq, msa_dir, SMILES=SMILES, CCD=CCD)

    all_jobs = df.apply(build_job, axis=1).tolist()

    if save_json:
        save_path = Path(save_json)
        save_path.parent.mkdir(parents=True, exist_ok=True)
        with save_path.open("w") as f:
            json.dump(all_jobs, f, indent=4)
        print(f"JSON saved to {save_path}")

    return all_jobs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/af3/02_protein_pairs.ipynb.

# %% auto 0
__all__ = ['get_colabfold_cmd', 'copy_a3m', 'a3m_to_seq', 'get_protein_subjson', 'dump_json_folder', 'get_multi_protein_json',
           'generate_pair_df']

# %% ../../nbs/af3/02_protein_pairs.ipynb 4
import os, json, shutil, pandas as pd
from tqdm import tqdm
from itertools import combinations
from pathlib import Path
from .json import *
from .docker import *

# %% ../../nbs/af3/02_protein_pairs.ipynb 9
def get_colabfold_cmd(csv_path,project_name):
    print('Run below in terminal:')
    print(f'\n colabfold_batch {csv_path} msa_{project_name} --msa-only')

# %% ../../nbs/af3/02_protein_pairs.ipynb 13
def copy_a3m(a3m_dir: str, # Path to the source directory containing .a3m files.
             dest_dir: str, # Path to the destination directory where files will be copied
             ):
    "Copies all .a3m files from the source directory to the destination directory."
    
    a3m_dir,dest_dir = Path(a3m_dir),Path(dest_dir)
    dest_dir.mkdir(parents=True, exist_ok=True)

    files = list(a3m_dir.glob('*.a3m'))

    for file in tqdm(files, desc="Copying files", unit="file"):
        shutil.copy(file, dest_dir / file.name)

    print(f"Copied {len(files)} a3m files from {a3m_dir} to {dest_dir}")

# %% ../../nbs/af3/02_protein_pairs.ipynb 17
def a3m_to_seq(file_path:Path):
    "Get protein sequence from a3m file"
    return file_path.read_text().splitlines()[2] # protein sequence is located on line 2

# %% ../../nbs/af3/02_protein_pairs.ipynb 19
def get_protein_subjson(gene_name, a3m_dir=".",idx = 'A',run_template=True):
    "Get subjson (protein part) with colabfold unpairedMSA .a3m path"
    file_path = Path(a3m_dir)/f"{gene_name}.a3m"
    protein_sequence = a3m_to_seq(file_path)
    
    json_data = {
        'id': idx,
        'sequence': protein_sequence, 
        'modifications': [],
        'unpairedMsaPath': str("/root"/file_path), # for docker path, ECD under af_input
        'pairedMsa': '',
        'templates': None if run_template else []
    }

    return json_data

# %% ../../nbs/af3/02_protein_pairs.ipynb 22
def dump_json_folder(json_data, folder):
    "Save json under a folder"
    file_path = Path(folder)/f"{json_data['name']}.json"
    with open(file_path,'w') as f: json.dump(json_data,f,indent=4)

# %% ../../nbs/af3/02_protein_pairs.ipynb 23
def get_multi_protein_json(gene_list,a3m_dir,run_template=True,save_folder=None):
    'Get json of multiple proteins, with unpaired MSA path indicated (from colabfold MSA)'
    sequences = []
    alphabets = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    for index, gene in enumerate(gene_list):
        sub_json=get_protein_subjson(gene,a3m_dir,idx=alphabets[index],run_template=run_template)
        sequences.append({'protein':sub_json})
    name = '_'.join(gene_list)
    json_data = {
            "name": name,
            "modelSeeds": [1],
            "sequences": sequences,
            "bondedAtomPairs": [],
            "dialect": "alphafold3",
            "version": 2
        }
    if save_folder:
        dump_json_folder(json_data,save_folder)
    return json_data

# %% ../../nbs/af3/02_protein_pairs.ipynb 27
def generate_pair_df(gene_list,self_pair=True):
    "Unique pair genes in a gene list"
    pairs = list(combinations(gene_list, 2))
    pair_df = pd.DataFrame(pairs,columns=["Gene1", "Gene2"])
    
    if self_pair:
        self_pair_df = pd.DataFrame({'Gene1':gene_list, 'Gene2':gene_list})
        pair_df = pd.concat([pair_df,self_pair_df])

    return pair_df.reset_index(drop=True)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/af3/03_analyze.ipynb.

# %% auto 0
__all__ = ['read_summary_json', 'get_summary_df', 'process_summary_df', 'get_top_cases', 'get_3d_report', 'get_report',
           'copy_file']

# %% ../../nbs/af3/03_analyze.ipynb 3
import json, shutil, pandas as pd
from pathlib import Path
import plotly.graph_objects as go

# %% ../../nbs/af3/03_analyze.ipynb 5
def read_summary_json(json_path):
    "Read json file to dictionary"
    json_path = Path(json_path)
    json_data=json.loads(json_path.read_text())
        
    # Flatten the JSON data into a single row
    row = {"ID": json_path.stem}
    for key, value in json_data.items():
        if isinstance(value, list):
            for idx, sub_value in enumerate(value):
                if isinstance(sub_value, list):
                    for sub_idx, sub_sub_value in enumerate(sub_value):
                        row[f"{key}_{idx}_{sub_idx}"] = sub_sub_value
                else:
                    row[f"{key}_{idx}"] = sub_value
        else:
            row[key] = value
    return row

# %% ../../nbs/af3/03_analyze.ipynb 7
def get_summary_df(output_dir):
    "Pack the summary json from the output folder to the df"
    
    path_list = list(Path(output_dir).rglob('*_summary_confidences.json'))
    print(len(path_list),'summary_confidences.json files detected')
    return pd.DataFrame(list(map(read_summary_json,path_list)))

# %% ../../nbs/af3/03_analyze.ipynb 10
def process_summary_df(df,generate_report=False):
    "Post process the json-converted pandas df; remove redundant columns; available for pairs"
    
    df['ID'] = df.ID.str.replace('_summary_confidences','')
    df = df.set_index('ID')

    # drop zero std columns: usually contains chain_pair_pae_min_0_0 and has_clash
    zero_std_cols = df.columns[df.std()<1e-10]
    df=df.drop(columns=zero_std_cols)

    # drop columns with same values
    redundant_columns = []
    
    # Check for columns equal to `iptm`
    for col in ['chain_iptm_0', 'chain_iptm_1', 'chain_pair_iptm_0_1', 'chain_pair_iptm_1_0']:
        if df['iptm'].equals(df[col]):
            redundant_columns.append(col)
    
    # Check for columns equal to `chain_ptm_0`
    if df['chain_ptm_0'].equals(df['chain_pair_iptm_0_0']):
        redundant_columns.append('chain_pair_iptm_0_0')
    
    # Check for columns equal to `chain_ptm_1`
    if df['chain_ptm_1'].equals(df['chain_pair_iptm_1_1']):
        redundant_columns.append('chain_pair_iptm_1_1')
    
    # Drop redundant columns
    df = df.drop(columns=redundant_columns)
    
    if generate_report:
        print('Generating pairplot graph report')
        sns.pairplot(data=df, corner=True)
        plt.savefig("af_report.pdf")
        plt.close()
        print('Export to af_report.pdf')

    # add iptm and ptm
    df['iptm_ptm_add'] = df['iptm']+df['ptm']

    # inter error add
    df['chain_pair_pae_min_add'] = df['chain_pair_pae_min_0_1']+df['chain_pair_pae_min_1_0']

    # rank iptm and ptm and add the rank, this method can ignore the different value range between iptm and ptm
    df['iptm_rnk'],df['ptm_rnk'] = df.iptm.rank(ascending=False), df.ptm.rank(ascending=False)
    df['iptm_ptm_rnk_add'] = df['iptm_rnk']+df['ptm_rnk']

    # combine chain_pair_pae_min & iptm rank
    df['chain_pair_pae_min_add_rnk'] = df['chain_pair_pae_min_add'].rank()
    df['iptm_pae_add_rnk'] = df['chain_pair_pae_min_add_rnk'] + df['iptm_rnk']
    
    return df

# %% ../../nbs/af3/03_analyze.ipynb 12
def get_top_cases(df,n=30):
    "Get top cases from the metric"
    idxs = set()
    
    big_cols = ['ranking_score', 'iptm', 'iptm_ptm_add']
    small_cols = [
        'iptm_ptm_rnk_add', 
        'chain_pair_pae_min_add', 
        'chain_pair_pae_min_0_1', 
        'chain_pair_pae_min_1_0', 
        'iptm_pae_add_rnk'
    ]
    
    for col in big_cols: idxs.update(df.nlargest(n, col).index)
    for col in small_cols: idxs.update(df.nsmallest(n, col).index)

    return list(idxs)

# %% ../../nbs/af3/03_analyze.ipynb 14
def get_3d_report(df,index_list, x='iptm',y='ptm',z='chain_pair_pae_min_add',save_dir='af_report'):
    "Generate 3d plot html file given case index and x, y, z colname"
    annotation = df.index.where(df.index.isin(index_list),'').str.split('_').str[1]

    # Create the 3D scatter plot
    fig = go.Figure(data=go.Scatter3d(
        x=df[x],
        y=df[y],
        z=df[z],
        mode='markers+text',
        text=annotation,  # Annotation using the index
        textposition="top center",
        marker=dict(size=8, color='blue', opacity=0.8),
    ))
    
    # Customize layout
    fig.update_layout(
        scene=dict(
            xaxis_title=x,
            yaxis_title=y,
            zaxis_title=z,
        ),
        title='3D Scatter Plot',
        autosize=True,
        height=3000,
    )
    
    fig.write_html(Path(save_dir)/'3d_scatter_plot.html',full_html=True)
    print('Exported the html to 3d_scatter_plot.html')

# %% ../../nbs/af3/03_analyze.ipynb 16
def get_report(out_dir,save_dir='af_report'):
    "Generate summary report based on summary_confidences file; return summary df and top cases"
    out = get_summary_df(out_dir)
    out = process_summary_df(out)
    top_cases = get_top_cases(out)
    
    Path(save_dir).mkdir(exist_ok=True)
    get_3d_report(out,top_cases,save_dir=save_dir)
    out.to_csv(Path(save_dir)/'summary_confidences.csv')
    
    return out, top_cases

# %% ../../nbs/af3/03_analyze.ipynb 18
def copy_file(idx_name, source_dir, dest_dir):
    "Copy all model cif generated by AF3 to the new dest folder"
    source_path = Path(source_dir)/f"{idx_name}/{idx_name}_model.cif"
    dest_path = Path(dest_dir)/f"{idx_name}_model.cif"
    shutil.copy(source_path,dest_path)
    print(f'Copying {str(source_path)} to {str(dest_path)}')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/af3/00_json.ipynb.

# %% auto 0
__all__ = ['dump_json', 'get_protein_json', 'read_json', 'get_protein_smiles_json', 'get_protein_ccdcode_json',
           'assign_atom_names_from_graph', 'mol_to_ccd_text', 'sdf2ccd', 'get_protein_ccd_json', 'split_nfolder']

# %% ../../nbs/af3/00_json.ipynb 2
import re, shutil, json, pandas as pd, numpy as np
from pathlib import Path

from rdkit import Chem as rd_chem
from rdkit.Chem import AllChem,rdmolfiles
from rdkit import Chem

from Bio.PDB import PDBParser

# %% ../../nbs/af3/00_json.ipynb 4
def dump_json(data, save_path):
    "Save json data into a file"
    with open(save_path,'w') as f: 
        json.dump(data,f,indent=4)

# %% ../../nbs/af3/00_json.ipynb 5
def get_protein_json(name, # job name
                     seq, # aa sequence
                     save_path=None, # .json
                     seeds=[1]
                     ):
    "Generate json of single protein sequence for input of docker command"
    
    json_data = {
        "name": name,
        "modelSeeds": seeds,
        "sequences": [
            {
                "protein": {
                    "id": "A",
                    "sequence": seq,
                }
            },
        ],
        "bondedAtomPairs": [],
        "dialect": "alphafold3",
        "version": 3
    }
    if save_path:
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        dump_json(json_data,save_path)
    return json_data

# %% ../../nbs/af3/00_json.ipynb 9
def read_json(file_path):
    with open(file_path,'r') as f: 
        data = json.load(f)
    return data

# %% ../../nbs/af3/00_json.ipynb 11
def get_protein_smiles_json(smi_id:str, 
                            SMILES:str, 
                            protein_json, # json type
                            save_path=None, # .json
                            seeds=[1]
                            ):
    
    "Get json for protein-ligand docking task"
    raw_smiles = r"{}".format(SMILES) # JSON escaping, \ to \\
    protein_index = next(i for i, item in enumerate(protein_json["sequences"]) if "protein" in item)
    json_data = {
        "name": smi_id,
        "modelSeeds": seeds,
        "sequences": [
            {
                "ligand": {
                    "id": "L",
                    "smiles": raw_smiles,
                }
            }, 
            {
                "protein": protein_json["sequences"][protein_index]["protein"]
            },
        ],
        "bondedAtomPairs": [],
        "dialect": "alphafold3",
        "version": 2
    }
    if save_path:
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        dump_json(json_data,save_path)
    return json_data

# %% ../../nbs/af3/00_json.ipynb 18
def get_protein_ccdcode_json(protein_json,  # dict with protein sequence
                              ccd_code,      # str or list of str
                              job_id: str,   # job/task ID
                              save_path=None,  # optional output path
                              seeds=[1]):      # optional random seeds
    "Create AlphaFold3 docking JSON with CCD code(s)."
    
    # Normalize ccd_code to a list
    if isinstance(ccd_code, str):
        ccd_code = [ccd_code]
    elif not isinstance(ccd_code, list):
        raise TypeError("ccd_code must be a string or a list of strings.")

    protein_index = next(i for i, item in enumerate(protein_json["sequences"]) if "protein" in item)

    json_data = {
        "name": job_id,
        "modelSeeds": seeds,
        "sequences": [
            {
                "ligand": {
                    "id": "L",
                    "ccdCodes": ccd_code
                }
            },
            {
                "protein": protein_json["sequences"][protein_index]["protein"]
            },
        ],
        "dialect": "alphafold3",
        "version": 3
    }

    if save_path:
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        dump_json(json_data, save_path)

    return json_data

# %% ../../nbs/af3/00_json.ipynb 22
# Mapping bond types to mmCIF-compatible values
_RDKIT_BOND_TYPE_TO_MMCIF = {
    rd_chem.BondType.SINGLE: 'SING',
    rd_chem.BondType.DOUBLE: 'DOUB',
    rd_chem.BondType.TRIPLE: 'TRIP',
    rd_chem.BondType.AROMATIC: 'AROM'
}

def assign_atom_names_from_graph(mol):
    for i, atom in enumerate(mol.GetAtoms()):
        atom.SetProp('atom_name', f"{atom.GetSymbol()}{i+1}")
    return mol

def mol_to_ccd_text(mol, component_id, pdbx_smiles=None, include_hydrogens=False):
    mol = rd_chem.Mol(mol)
    if include_hydrogens:
        mol = rd_chem.AddHs(mol)
    rd_chem.Kekulize(mol, clearAromaticFlags=True)

    if mol.GetNumConformers() == 0:
        raise ValueError('The molecule has no conformers')
    conf = mol.GetConformer()
    coords = conf.GetPositions()

    mol = assign_atom_names_from_graph(mol)
    atom_map = {atom.GetIdx(): atom.GetProp('atom_name') for atom in mol.GetAtoms()}

    lines = [
        f"data_{component_id}",
        "#",
        f"_chem_comp.id {component_id}",
        f"_chem_comp.name '{component_id}'",
        "_chem_comp.type non-polymer",
        "_chem_comp.formula '?'",
        "_chem_comp.mon_nstd_parent_comp_id ?",
        "_chem_comp.pdbx_synonyms ?",
        "_chem_comp.formula_weight '?'",
    ]
    if pdbx_smiles:
        lines.append(f"_chem_comp.pdbx_smiles {pdbx_smiles}")
    lines += [
        "#",
        "loop_",
        "_chem_comp_atom.comp_id",
        "_chem_comp_atom.atom_id",
        "_chem_comp_atom.type_symbol",
        "_chem_comp_atom.charge",
        "_chem_comp_atom.pdbx_leaving_atom_flag",
        "_chem_comp_atom.pdbx_model_Cartn_x_ideal",
        "_chem_comp_atom.pdbx_model_Cartn_y_ideal",
        "_chem_comp_atom.pdbx_model_Cartn_z_ideal"
    ]

    for i, atom in enumerate(mol.GetAtoms()):
        if not include_hydrogens and atom.GetSymbol() == 'H':
            continue
        x, y, z = coords[i]
        lines.append(f"{component_id} {atom_map[atom.GetIdx()]} {atom.GetSymbol()} {atom.GetFormalCharge()} N {x:.3f} {y:.3f} {z:.3f}")

    lines += [
        "#",
        "loop_",
        "_chem_comp_bond.atom_id_1",
        "_chem_comp_bond.atom_id_2",
        "_chem_comp_bond.value_order",
        "_chem_comp_bond.pdbx_aromatic_flag"
    ]

    for bond in mol.GetBonds():
        a1, a2 = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
        if not include_hydrogens and (mol.GetAtomWithIdx(a1).GetSymbol() == 'H' or mol.GetAtomWithIdx(a2).GetSymbol() == 'H'):
            continue
        bond_type = _RDKIT_BOND_TYPE_TO_MMCIF[bond.GetBondType()]
        aromatic_flag = 'Y' if bond.GetIsAromatic() else 'N'
        lines.append(f"{atom_map[a1]} {atom_map[a2]} {bond_type} {aromatic_flag}")
    lines.append("#")

    return "\n".join(lines)

# %% ../../nbs/af3/00_json.ipynb 23
def sdf2ccd(sdf_path,
            CCD_name='lig-1', # do not use '_'; use as less letter as possible, 'lig-any' leads to extra ligands
            ):

    "Convert the compound to the AF3 required CCD format"
    supplier = Chem.SDMolSupplier(sdf_path)
    mol = supplier[0]  # Get the first molecule
    return mol_to_ccd_text(mol,CCD_name)

# %% ../../nbs/af3/00_json.ipynb 26
def get_protein_ccd_json(protein_json, # dict with protein sequence
                         rec_residue_num:int, # 1-indexed, for bondedAtomPairs, e.g., ["A", 145, "SG"] 
                         rec_atom_id:str, # for bondedAtomPairs, e.g., ["A", 145, "SG"] 
                         lig_sdf_path, # ccd text
                         lig_atom_id:str, # 0-indexed, for bondedAtomPairs, ["L", 1, "C04"]
                         job_id:str, # str, job/task ID
                         save_path=None,# optional output path
                         seeds=[1], # optional random seeds
                         ):               
    "Create AlphaFold3 docking JSON with customized CCD ligand and bondedAtomPairs."

    # get userCCD
    userCCD=sdf2ccd(lig_sdf_path)
    ccd_id = re.search(r"_chem_comp.id\s+([^\s#]+)", userCCD).group(1)
    
    protein_index = next(i for i, item in enumerate(protein_json["sequences"]) if "protein" in item)

    json_data = {
        "name": job_id,
        "modelSeeds": seeds,
        "sequences": [
            {
                "ligand": {
                    "id": "L",
                    "ccdCodes": [ccd_id]
                }
            },
            {
                "protein": protein_json["sequences"][protein_index]["protein"]
            },
        ],
        "bondedAtomPairs": [[["A", rec_residue_num, rec_atom_id],["L", 1, lig_atom_id]]],
        "userCCD": userCCD,
        "dialect": "alphafold3",
        "version": 3
    }

    if save_path:
        Path(save_path).parent.mkdir(parents=True, exist_ok=True)
        dump_json(json_data, save_path)

    return json_data

# %% ../../nbs/af3/00_json.ipynb 30
def split_nfolder(folder_dir, 
                  n=4):
    "Move json files from a folder into subfolders (folder_0, folder_1, ..., folder_N)."
    
    folder_dir = Path(folder_dir)

    files = sorted(folder_dir.glob("*.json"))
    # print(len(files))
    subfolders = [folder_dir / f"folder_{i}" for i in range(n)]
    for folder in subfolders:
        folder.mkdir(exist_ok=True)

    for idx, file in enumerate(files):
        target_folder = subfolders[idx % n]
        shutil.move(str(file), target_folder / file.name)

    print(f"Distributed {len(files)} files into {n} folders.")



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/kdock',
                'doc_host': 'https://sky1ove.github.io',
                'git_url': 'https://github.com/sky1ove/kdock',
                'lib_path': 'kdock'},
  'syms': { 'kdock.af3.analyze': { 'kdock.af3.analyze.copy_file': ('af3/analyze.html#copy_file', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.get_3d_report': ('af3/analyze.html#get_3d_report', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.get_report': ('af3/analyze.html#get_report', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.get_summary_df': ('af3/analyze.html#get_summary_df', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.get_top_cases': ('af3/analyze.html#get_top_cases', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.process_summary_df': ('af3/analyze.html#process_summary_df', 'kdock/af3/analyze.py'),
                                   'kdock.af3.analyze.read_summary_json': ('af3/analyze.html#read_summary_json', 'kdock/af3/analyze.py')},
            'kdock.af3.docker': { 'kdock.af3.docker.docker_multi_full': ('af3/docker.html#docker_multi_full', 'kdock/af3/docker.py'),
                                  'kdock.af3.docker.docker_multi_infer': ('af3/docker.html#docker_multi_infer', 'kdock/af3/docker.py'),
                                  'kdock.af3.docker.docker_multi_msa': ('af3/docker.html#docker_multi_msa', 'kdock/af3/docker.py'),
                                  'kdock.af3.docker.docker_single_full': ('af3/docker.html#docker_single_full', 'kdock/af3/docker.py'),
                                  'kdock.af3.docker.get_docker_command': ('af3/docker.html#get_docker_command', 'kdock/af3/docker.py')},
            'kdock.af3.json': { 'kdock.af3.json.assign_atom_names_from_graph': ( 'af3/json.html#assign_atom_names_from_graph',
                                                                                 'kdock/af3/json.py'),
                                'kdock.af3.json.dump_json': ('af3/json.html#dump_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.get_protein_ccd_json': ('af3/json.html#get_protein_ccd_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.get_protein_ccdcode_json': ('af3/json.html#get_protein_ccdcode_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.get_protein_json': ('af3/json.html#get_protein_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.get_protein_smiles_json': ('af3/json.html#get_protein_smiles_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.mol_to_ccd_text': ('af3/json.html#mol_to_ccd_text', 'kdock/af3/json.py'),
                                'kdock.af3.json.read_json': ('af3/json.html#read_json', 'kdock/af3/json.py'),
                                'kdock.af3.json.sdf2ccd': ('af3/json.html#sdf2ccd', 'kdock/af3/json.py'),
                                'kdock.af3.json.split_nfolder': ('af3/json.html#split_nfolder', 'kdock/af3/json.py')},
            'kdock.af3.protein_pairs': { 'kdock.af3.protein_pairs.a3m_to_seq': ( 'af3/protein_pairs.html#a3m_to_seq',
                                                                                 'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.copy_a3m': ( 'af3/protein_pairs.html#copy_a3m',
                                                                               'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.dump_json_folder': ( 'af3/protein_pairs.html#dump_json_folder',
                                                                                       'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.generate_pair_df': ( 'af3/protein_pairs.html#generate_pair_df',
                                                                                       'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.get_colabfold_cmd': ( 'af3/protein_pairs.html#get_colabfold_cmd',
                                                                                        'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.get_multi_protein_json': ( 'af3/protein_pairs.html#get_multi_protein_json',
                                                                                             'kdock/af3/protein_pairs.py'),
                                         'kdock.af3.protein_pairs.get_protein_subjson': ( 'af3/protein_pairs.html#get_protein_subjson',
                                                                                          'kdock/af3/protein_pairs.py')},
            'kdock.core.data': { 'kdock.core.data.Collins': ('core/data.html#collins', 'kdock/core/data.py'),
                                 'kdock.core.data.Collins.get_antibiotics_2k': ( 'core/data.html#collins.get_antibiotics_2k',
                                                                                 'kdock/core/data.py'),
                                 'kdock.core.data.Collins.get_antibiotics_39k': ( 'core/data.html#collins.get_antibiotics_39k',
                                                                                  'kdock/core/data.py'),
                                 'kdock.core.data.Collins.get_antibiotics_enzyme': ( 'core/data.html#collins.get_antibiotics_enzyme',
                                                                                     'kdock/core/data.py'),
                                 'kdock.core.data.Kras': ('core/data.html#kras', 'kdock/core/data.py'),
                                 'kdock.core.data.Kras.get_mirati_g12d': ('core/data.html#kras.get_mirati_g12d', 'kdock/core/data.py'),
                                 'kdock.core.data.Kras.get_mirati_g12d_raw': ( 'core/data.html#kras.get_mirati_g12d_raw',
                                                                               'kdock/core/data.py'),
                                 'kdock.core.data.Kras.get_seq': ('core/data.html#kras.get_seq', 'kdock/core/data.py'),
                                 'kdock.core.data.fetch_csv': ('core/data.html#fetch_csv', 'kdock/core/data.py')},
            'kdock.core.ligand': { 'kdock.core.ligand.compress_fp': ('core/ligand.html#compress_fp', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_fp': ('core/ligand.html#get_fp', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_rdkit': ('core/ligand.html#get_rdkit', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_rdkit_3d': ('core/ligand.html#get_rdkit_3d', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_rdkit_all': ('core/ligand.html#get_rdkit_all', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_rdkit_df': ('core/ligand.html#get_rdkit_df', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.get_same_mol_group': ('core/ligand.html#get_same_mol_group', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.hash_fp': ('core/ligand.html#hash_fp', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.plot_drug': ('core/ligand.html#plot_drug', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.preprocess': ('core/ligand.html#preprocess', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.rdkit_conformer': ('core/ligand.html#rdkit_conformer', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.remove_hi_corr': ('core/ligand.html#remove_hi_corr', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.tanimoto': ('core/ligand.html#tanimoto', 'kdock/core/ligand.py'),
                                   'kdock.core.ligand.tanimoto_numba': ('core/ligand.html#tanimoto_numba', 'kdock/core/ligand.py')},
            'kdock.core.plot': { 'kdock.core.plot.plot_2d': ('core/plot.html#plot_2d', 'kdock/core/plot.py'),
                                 'kdock.core.plot.plot_corr': ('core/plot.html#plot_corr', 'kdock/core/plot.py'),
                                 'kdock.core.plot.reduce_feature': ('core/plot.html#reduce_feature', 'kdock/core/plot.py'),
                                 'kdock.core.plot.set_sns': ('core/plot.html#set_sns', 'kdock/core/plot.py')},
            'kdock.core.protein': { 'kdock.core.protein.apply_mut_complex': ( 'core/protein.html#apply_mut_complex',
                                                                              'kdock/core/protein.py'),
                                    'kdock.core.protein.apply_mut_single': ('core/protein.html#apply_mut_single', 'kdock/core/protein.py'),
                                    'kdock.core.protein.compare_seq': ('core/protein.html#compare_seq', 'kdock/core/protein.py'),
                                    'kdock.core.protein.get_uniprot_features': ( 'core/protein.html#get_uniprot_features',
                                                                                 'kdock/core/protein.py'),
                                    'kdock.core.protein.get_uniprot_kd': ('core/protein.html#get_uniprot_kd', 'kdock/core/protein.py'),
                                    'kdock.core.protein.get_uniprot_seq': ('core/protein.html#get_uniprot_seq', 'kdock/core/protein.py'),
                                    'kdock.core.protein.get_uniprot_type': ('core/protein.html#get_uniprot_type', 'kdock/core/protein.py')},
            'kdock.core.utils': { 'kdock.core.utils.copy_files': ('core/utils.html#copy_files', 'kdock/core/utils.py'),
                                  'kdock.core.utils.get_box': ('core/utils.html#get_box', 'kdock/core/utils.py'),
                                  'kdock.core.utils.get_rec_lig': ('core/utils.html#get_rec_lig', 'kdock/core/utils.py'),
                                  'kdock.core.utils.rglob': ('core/utils.html#rglob', 'kdock/core/utils.py'),
                                  'kdock.core.utils.view_complex': ('core/utils.html#view_complex', 'kdock/core/utils.py'),
                                  'kdock.core.utils.view_mol': ('core/utils.html#view_mol', 'kdock/core/utils.py')},
            'kdock.gnina.dock': { 'kdock.gnina.dock.extract_gnina_dock': ( 'gnina/gnina_docking.html#extract_gnina_dock',
                                                                           'kdock/gnina/dock.py'),
                                  'kdock.gnina.dock.gnina_dock': ('gnina/gnina_docking.html#gnina_dock', 'kdock/gnina/dock.py'),
                                  'kdock.gnina.dock.setup_gnina_docker': ( 'gnina/gnina_docking.html#setup_gnina_docker',
                                                                           'kdock/gnina/dock.py'),
                                  'kdock.gnina.dock.setup_gnina_local': ( 'gnina/gnina_docking.html#setup_gnina_local',
                                                                          'kdock/gnina/dock.py')},
            'kdock.gnina.rescore': { 'kdock.gnina.rescore.ChainSelect': ( 'gnina/gnina_af3_rescore.html#chainselect',
                                                                          'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.ChainSelect.__init__': ( 'gnina/gnina_af3_rescore.html#chainselect.__init__',
                                                                                   'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.ChainSelect.accept_chain': ( 'gnina/gnina_af3_rescore.html#chainselect.accept_chain',
                                                                                       'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.extract_gnina_rescore': ( 'gnina/gnina_af3_rescore.html#extract_gnina_rescore',
                                                                                    'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.get_gnina_rescore': ( 'gnina/gnina_af3_rescore.html#get_gnina_rescore',
                                                                                'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.get_gnina_rescore_folder': ( 'gnina/gnina_af3_rescore.html#get_gnina_rescore_folder',
                                                                                       'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.gnina_rescore_docker': ( 'gnina/gnina_af3_rescore.html#gnina_rescore_docker',
                                                                                   'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.gnina_rescore_local': ( 'gnina/gnina_af3_rescore.html#gnina_rescore_local',
                                                                                  'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.pdb2sdf': ('gnina/gnina_af3_rescore.html#pdb2sdf', 'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.prepare_rec_lig': ( 'gnina/gnina_af3_rescore.html#prepare_rec_lig',
                                                                              'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.rename_residues': ( 'gnina/gnina_af3_rescore.html#rename_residues',
                                                                              'kdock/gnina/rescore.py'),
                                     'kdock.gnina.rescore.split_cif': ('gnina/gnina_af3_rescore.html#split_cif', 'kdock/gnina/rescore.py')},
            'kdock.px.core': { 'kdock.px.core.get_protein_ligand_df_json': ( 'protenix/protenix.html#get_protein_ligand_df_json',
                                                                             'kdock/px/core.py'),
                               'kdock.px.core.get_single_job': ('protenix/protenix.html#get_single_job', 'kdock/px/core.py'),
                               'kdock.px.core.get_single_protein_ligand_json': ( 'protenix/protenix.html#get_single_protein_ligand_json',
                                                                                 'kdock/px/core.py'),
                               'kdock.px.core.get_virtual_screening_json': ( 'protenix/protenix.html#get_virtual_screening_json',
                                                                             'kdock/px/core.py')},
            'kdock.px.dock': { 'kdock.px.dock.capture_output': ('protenix/proteinix_dock.html#capture_output', 'kdock/px/dock.py'),
                               'kdock.px.dock.get_protenix_dock': ('protenix/proteinix_dock.html#get_protenix_dock', 'kdock/px/dock.py'),
                               'kdock.px.dock.get_protenix_vina_dock': ( 'protenix/proteinix_dock.html#get_protenix_vina_dock',
                                                                         'kdock/px/dock.py'),
                               'kdock.px.dock.json2sdf': ('protenix/proteinix_dock.html#json2sdf', 'kdock/px/dock.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/gnina/04_gnina_docking.ipynb.

# %% auto 0
__all__ = ['setup_gnina_local', 'setup_gnina_docker', 'extract_gnina_dock', 'gnina_dock']

# %% ../../nbs/gnina/04_gnina_docking.ipynb 3
# basics
import re,subprocess, py3Dmol
from tqdm import tqdm
from pathlib import Path
import pandas as pd,numpy as np

# rdkit
from rdkit import Chem
from rdkit.Chem import AllChem

# %% ../../nbs/gnina/04_gnina_docking.ipynb 8
def setup_gnina_local(version='v1.3'):
    "Download and install gnina in the current directory"
    # Check CUDA availability
    # try:
    #     subprocess.run(["nvidia-smi"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    # except subprocess.CalledProcessError:
    #     raise EnvironmentError("CUDA not detected. Please make sure a CUDA-capable GPU is available and drivers are installed.")
    # except FileNotFoundError:
    #     raise EnvironmentError("nvidia-smi not found. Make sure NVIDIA drivers and CUDA are installed.")

    subprocess.run(["sudo", "apt-get", "update", "-yq"], check=True,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    subprocess.run(["sudo", "apt-get", "install", "-yq", "openbabel"], check=True)

    gnina_url = f"https://github.com/gnina/gnina/releases/download/{version}/gnina"
    print(f'Downloading {version} gnina')
    subprocess.run(["wget",gnina_url], check=True,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    subprocess.run(["chmod", "+x", 'gnina'], check=True)
    
    print('Finish setup!')

# %% ../../nbs/gnina/04_gnina_docking.ipynb 13
def setup_gnina_docker():
    "Pull gnina docker image"
    print("Pulling GNINA Docker image: gnina/gnina")
    subprocess.run(["docker", "pull", "gnina/gnina"], check=True,stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    print("GNINA Docker image is ready.")

# %% ../../nbs/gnina/04_gnina_docking.ipynb 18
def extract_gnina_dock(gnina_output):
    "Extract values from gnina output"
    mode1_line = re.search(r'\b1\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\s+(-?\d+\.\d+)\b', gnina_output)
    
    if mode1_line:
        affinity = float(mode1_line.group(1))
        cnn_pose_score = float(mode1_line.group(2))
        cnn_affinity = float(mode1_line.group(3))
        
        return affinity, cnn_pose_score, cnn_affinity
    
    return None

# %% ../../nbs/gnina/04_gnina_docking.ipynb 19
def gnina_dock(receptor, # receptor file
              ligand, # ligand file
              autobox_ligand, # ligand file isolated from the complex
              output = 'docked.sdf', # output file (sdf or sdf.gz) to be saved
              seed=0, # set seeds
              exhaustiveness=None, # number of MC chains, default is 8 if None, the higher the better (16,32); for whole protein, use 64
              ):
    
    command = ['./gnina', 
               '-r', receptor, 
               '-l', ligand, 
               '--autobox_ligand', autobox_ligand,
               '-o', output,
               '--seed', str(seed)]

    if exhaustiveness is not None:
        command.extend(['--exhaustiveness', str(exhaustiveness)])

    output_txt = subprocess.run(command, capture_output=True, text=True).stdout
    
    print(f'save the docked file as {output}')
    
    values = extract_gnina_dock(output_txt)
    
    print(f'affinity, cnn_pose_score, and cnn_affinity are: {values}')

    return values

# %% ../../nbs/gnina/04_gnina_docking.ipynb 21
def gnina_dock(df, 
                 ID_col = 'ID',
                 smi_col = 'SMILES',
              output_dir = 'gnina_docked'
                 ):
    affinity_values = []
    cnn_pose_score_values = []
    cnn_affinity_values = []

    
    Path(output_dir).mkdir(parents=True,exist_ok=True)

    for i, r in tqdm(df.iterrows(),total=len(df),desc='Docking'):
        rdkit_conformer(SMILES=r[smi_col], output = f'ligand/{r[ID_col]}.sdf', visualize=False)
        affinity, cnn_pose_score, cnn_affinity = gnina_dock('rec.pdb',f'ligand/{r[ID_col]}.sdf', 'lig.pdb',f'docked/docked_{r[ID_col]}.sdf')

        affinity_values.append(affinity)
        cnn_pose_score_values.append(cnn_pose_score)
        cnn_affinity_values.append(cnn_affinity)

    df = df.copy()
    df['Affinity'] = affinity_values
    df['CNN_Pose_Score'] = cnn_pose_score_values
    df['CNN_Affinity'] = cnn_affinity_values
    
    return df



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/gnina/05_gnina_AF3_rescore.ipynb.

# %% auto 0
__all__ = ['ChainSelect', 'rename_residues', 'split_cif', 'pdb2sdf', 'prepare_rec_lig', 'gnina_rescore_local',
           'gnina_rescore_docker', 'extract_gnina_rescore', 'get_gnina_rescore', 'get_gnina_rescore_folder']

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 3
import pandas as pd
import re, os, subprocess, py3Dmol
from Bio.PDB import MMCIFParser, PDBIO, Select
from rdkit import Chem
from rdkit.Chem import AllChem
from pathlib import Path
from fastcore.all import L
from tqdm.contrib.concurrent import process_map
from functools import partial

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 5
class ChainSelect(Select):
    "Select chain to save"
    def __init__(self, chain_ids):
        self.chain_ids = chain_ids
    def accept_chain(self, chain):
        return chain.get_id() in self.chain_ids

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 6
def rename_residues(structure, chain_id, new_resname='LIG'):
    "Rename residue name from LIG_L to LIG as LIG_L exceeds lengths and leads to error in RDKit"
    for model in structure:
        for chain in model:
            if chain.id == chain_id:
                for residue in chain:
                    residue.resname = new_resname

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 7
def split_cif(cif_path, rec_chain_id,lig_chain_id, rec_pdb_path, lig_pdb_path):
    "Split AF3 output CIF to protein and ligand PDBs"
    parser = MMCIFParser(QUIET=True)
    structure = parser.get_structure('complex', cif_path)
    rename_residues(structure, chain_id=lig_chain_id, new_resname='LIG')
    io = PDBIO()
    io.set_structure(structure)
    io.save(str(rec_pdb_path), ChainSelect(rec_chain_id))  # receptor
    io.save(str(lig_pdb_path), ChainSelect(lig_chain_id))  # ligand

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 8
def pdb2sdf(pdb_path, sdf_path):
    "Convert ligand pdb to sdf file"
    mol = Chem.MolFromPDBFile(pdb_path, sanitize=True, removeHs=False)
    if mol:
        writer = Chem.SDWriter(sdf_path)
        writer.write(mol)
        writer.close()
        return None
    else:
        print('Conversion failed for:', pdb_path)
        return pdb_path

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 9
def prepare_rec_lig(cif_path, rec_chain_id, lig_chain_id, rec_pdb_path, lig_pdb_path):
    "Split AF3 cif to protein.pdb (chainA) and ligand.sdf (chainL) "
    
    tmp_name = Path(cif_path).stem
    tmp_path = f'{tmp_name}_lig.pdb'
    split_cif(cif_path, rec_chain_id,lig_chain_id, rec_pdb_path, tmp_path)
    failed = pdb2sdf(tmp_path, lig_pdb_path)
    try:
        os.remove(tmp_path)
    except OSError:
        pass
    return failed

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 15
def gnina_rescore_local(protein_pdb,  # receptor file
                        ligand_sdf,   # ligand file
                        CNN_affinity=True,
                        vinardo=False, # if True, use vinardo instead of vina
                        ):

    command = ['./gnina', 
               '-r', protein_pdb, 
               '-l', ligand_sdf,
               '--minimize']  # always include this

    # Handle scoring options
    if not CNN_affinity:
        command += ['--cnn_scoring', 'none']
    if vinardo:
        command += ['--scoring', 'vinardo']

    result = subprocess.run(command, capture_output=True, text=True)
    return result.stdout

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 17
def gnina_rescore_docker(protein_pdb, 
                         ligand_sdf, 
                         CNN_affinity=True, 
                         vinardo=False):
    "Run GNINA rescoring using Docker. Supports receptor and ligand in different folders."

    protein_pdb = Path(protein_pdb).resolve()
    ligand_sdf = Path(ligand_sdf).resolve()

    # Mount points inside the Docker container
    rec_mount = '/recdata'
    lig_mount = '/ligdata'

    command = [
        'docker', 'run', '--rm',
        '-v', f'{protein_pdb.parent}:{rec_mount}',  # mount receptor directory
        '-v', f'{ligand_sdf.parent}:{lig_mount}',   # mount ligand directory
        'gnina/gnina',
        'gnina',
        '-r', f'{rec_mount}/{protein_pdb.name}',
        '-l', f'{lig_mount}/{ligand_sdf.name}',
        '--minimize',  # always include
    ]

    if not CNN_affinity:
        command += ['--cnn_scoring', 'none']
    if vinardo:
        command += ['--scoring', 'vinardo']

    result = subprocess.run(command, capture_output=True, text=True)
    return result.stdout

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 19
def extract_gnina_rescore(txt):
    """Extract GNINA output metrics into a dictionary (partial match allowed)."""
    result = {}

    patterns = {
        'binding_energy': r'Affinity:\s+([-.\d]+)',
        'uncertainty':    r'Affinity:\s+[-.\d]+\s+([-.\d]+)',
        'RMSD':           r'RMSD:\s+([-.\d]+)',
        'CNNscore':       r'CNNscore:\s+([-.\d]+)',
        'CNNaffinity':    r'CNNaffinity:\s+([-.\d]+)',
        'CNNvariance':    r'CNNvariance:\s+([-.\d]+)',
    }

    for key, pat in patterns.items():
        match = re.search(pat, txt)
        if match:
            result[key] = float(match.group(1))

    return result

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 24
def get_gnina_rescore(cif_path,
                      rec_chain_id='A', 
                      lig_chain_id='L', 
                      CNN_affinity=True,
                      vinardo=False,
                      is_local=True):
    "Split the CIF into receptor and ligand folders, then extract the GNINA rescored affinity score"
    cif_path = Path(cif_path).expanduser()
    parent,stem = cif_path.parent,cif_path.stem

    rec_dir,lig_dir = Path(str(parent) + '_receptor'),Path(str(parent) + '_ligand')
    
    rec_path,lig_path = rec_dir/f'{stem}.pdb',lig_dir/f'{stem}.sdf'
    
    rec_dir.mkdir(exist_ok=True)
    lig_dir.mkdir(exist_ok=True)
    
    prepare_rec_lig(cif_path,rec_chain_id, lig_chain_id,rec_path,lig_path)
    if is_local:
        gnina_output = gnina_rescore_local(rec_path,lig_path,CNN_affinity,vinardo)
    else:
        gnina_output = gnina_rescore_docker(rec_path,lig_path,CNN_affinity,vinardo)
    return extract_gnina_rescore(gnina_output)

# %% ../../nbs/gnina/05_gnina_AF3_rescore.ipynb 29
def get_gnina_rescore_folder(cif_folder,
                             rec_chain_id='A', 
                             lig_chain_id='L',
                             CNN_affinity=True,
                             vinardo=False,
                             is_local=True):
    "Parallel processing to get gnina rescore given folder path"
    cifs = L(Path(cif_folder).expanduser().glob("*.cif")) # just take cif file
    
    func = partial(get_gnina_rescore,
                   rec_chain_id=rec_chain_id, 
                   lig_chain_id=lig_chain_id,
                   CNN_affinity=CNN_affinity,
                   vinardo=vinardo,
                   is_local=is_local)
    results = process_map(func, cifs, max_workers=4)

    # use path.stem as df index
    results_dict = dict(zip([p.stem for p in cifs], results))
    result_df = pd.DataFrame(results_dict).T.reset_index(names='ID')
    
    prefix = "vinardo_" if vinardo else "vina_"
    result_df = result_df.rename(columns={
        "binding_energy": f"{prefix}binding_energy",
        "uncertainty": f"{prefix}uncertainty",
        "RMSD": f"{prefix}RMSD"
    })

    return result_df

    
